package database;
import java.io.IOException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import com.mysql.cj.jdbc.exceptions.MysqlDataTruncation;

import dump.DumpReader;
import main.Constants;

public class DatabaseInserter extends Thread {
	
	private DatabaseConnectionCreator databaseHandler;
	
	private DumpReader dumpReader;
	
	public DatabaseInserter(DatabaseConnectionCreator databaseHandler, DumpReader dumpReader) {
		this.databaseHandler = databaseHandler;
		this.dumpReader = dumpReader;
	}
	
	@Override
	public void run() {
		try {
			fillDatabase();
		} catch (IOException e) {
			System.err.println("Error! Dump file missing or can not be opened!");
			System.exit(1);
		}
	}
	
	private void fillDatabase() throws IOException {
		List<String> page;
		while((page = dumpReader.getNextPage()) != null) {
			String title = getTitle(page);
			if (title == null) {
				continue;
			}
			
			String redirect = getRedirect(page);
			if (redirect != null) {
				insertRedirectTable(title, redirect);
			}
						
			List<InfoboxEntry> infoboxes = getInfoboxes(page);
			
			if (infoboxes.size() == 0) {
				continue;
			}
			
			int id = insertToPagesTable(title);
			
			for (InfoboxEntry infobox: infoboxes) {				
				insertToInfoboxesTable(infobox.getInfoboxType(), infobox.getInfoboxContent(), id);
			}
		}
	}
	
	/**
	 * Gets title of the give page.
	 * @param page list of lines of page
	 * @return title of page
	 */
	private String getTitle(List<String> page) {
		for (String line: page) {
			if (line.contains(Constants.TITLE_START) && line.contains(Constants.TITLE_END)) {
				return line.substring(line.indexOf(Constants.TITLE_START) + Constants.TITLE_START.length(), line.lastIndexOf(Constants.TITLE_END));				
			}
		}
		return null;
	}
		
	/**
	 * It tries to get target of redirected page. If none is presented, returns null.
	 * @param page List of page lines.
	 * @return redirect target value or null
	 */
	private String getRedirect(List<String> page) {
		for (String line: page) {
			if (line.contains(Constants.REDIRECT_START) && line.contains(Constants.REDIRECT_END)) {
				return line.substring(line.indexOf("\"") + 1, line.lastIndexOf("\""));
			}
		}
		return null;
	}
	
	/**
	 * Gets index of start and end of next infobox in the page.
	 * @param pageString string of the page
	 * @return start and end infobox indexes
	 */
	private int[] getInfoboxIndexes(String pageString) {
		int startInfoboxIndex = pageString.indexOf(Constants.INFOBOX_START);
		
		if (startInfoboxIndex == -1) {
			return new int[]{-1, -1};
		}
		
		int brackets = 0;
		
		int endInfoboxIndex = pageString.length();
		
		for (int i = startInfoboxIndex; i < pageString.length(); i++) {
			if (pageString.charAt(i) == '{') {
				brackets++;
			}
			else if (pageString.charAt(i) == '}') {
				brackets--;
			}
			
			if (brackets == 0) {
				endInfoboxIndex = i;
				break;
			}
		}
		
		return new int[]{startInfoboxIndex, endInfoboxIndex};
	}
	
	/**
	 * Gets type of infobox type in given page.
	 * @param pageString string of whole page
	 * @return type of the infobox
	 */
	private String getInfoboxType(String pageString) {
		List<String> pageLines = Arrays.asList(pageString.split(System.lineSeparator()));
		for (String line: pageLines) {
			if (line.contains(Constants.INFOBOX_START)) {
				if (line.contains(Constants.INFOBOX_KEY_VALUE_SEPARATOR)){
					String[] lineSplitted = line.split(Constants.INFOBOX_KEY_VALUE_SEPARATOR_ESCAPED);
					for (String splittedPart: lineSplitted) {
						if (splittedPart.contains(Constants.INFOBOX_START)) {
							return splittedPart.substring(splittedPart.indexOf(Constants.INFOBOX_START) + Constants.INFOBOX_START.length());
						}
					}
				}
				else {
					return line.substring(line.indexOf(Constants.INFOBOX_START) + Constants.INFOBOX_START.length());									
				}
			}
		}
		return null;
	}
	
	/**
	 * Converts list of strings which represents lines into one string.
	 * @param stringsList list of strings (lines of text)
	 * @return united string of list
	 */
	private String stringsListToString(List<String> stringsList) {
		StringBuilder listString = new StringBuilder();
		for (String oneString: stringsList) {
			listString.append(oneString);
			listString.append(System.lineSeparator());
		}
		return listString.toString();
	}
	
	/**
	 * It gets content of the infobox using given infobox indexes.
	 * @param pageString string of the page
	 * @param infoboxIndexes indexes of the infobox
	 * @return content of the infobox; if start or end of the infobox was not found, returns null
	 */
	private String getInfoboxContent(String pageString, int[] infoboxIndexes) {
		if (infoboxIndexes[0] != -1 || infoboxIndexes[1] != -1) {
			return pageString.substring(infoboxIndexes[0], infoboxIndexes[1]);
		}
		return null;
	}
	
	private List<InfoboxEntry> getInfoboxes(List<String> page) {
		List<InfoboxEntry> infoboxes = new ArrayList<InfoboxEntry>();
		String pageString = stringsListToString(page);
		
		while (true) {
			int[] infoboxIndexes = getInfoboxIndexes(pageString);
			String infoboxType = getInfoboxType(pageString);
			String infoboxContent = getInfoboxContent(pageString, infoboxIndexes);
			if (infoboxType != null && infoboxContent != null) {
				InfoboxEntry foundInfobox = new InfoboxEntry(infoboxType, infoboxContent);
				infoboxes.add(foundInfobox);
			}
			else {
				break;
			}
			pageString = pageString.substring(infoboxIndexes[1]);
		}
		
		return infoboxes;
	}
	
	/**
	 * Inserts found page name with at least one infobox to the database.
	 * @param tittle page name
	 * @return id of inserted page
	 */
	private int insertToPagesTable(String tittle) {
		PreparedStatement st;
		int id = 0;
		try {
			String query = ("INSERT INTO `wikipedia`.`pages` (`pageName`) VALUES (?)");
			st = databaseHandler.getConnection().prepareStatement(query, Statement.RETURN_GENERATED_KEYS);
			
			st.setString(1, tittle);
			
			st.executeUpdate();
			
			ResultSet rs = st.getGeneratedKeys();
			
			if (rs.next()) {
				id = rs.getInt(1);
			}
			st.close();
		}  catch (SQLException e) {
			e.printStackTrace();
		}
		return id;
	}
	
	/**
	 * Inserts found infobox of the page to the database.
	 * @param infoboxType type of found infobox
	 * @param infoboxContent content of found infobox
	 * @param pageId id of page where infobox was found
	 */
	private void insertToInfoboxesTable(String infoboxType, String infoboxContent, int pageId) {
		PreparedStatement st;
		try {
			String query = ("INSERT INTO `wikipedia`.`infoboxes`  (`infoboxType`, `infoboxContent`, `pageId`) VALUES (?, ?, ?)");
			st = databaseHandler.getConnection().prepareStatement(query, Statement.RETURN_GENERATED_KEYS);
			
			st.setString(1, infoboxType);
			st.setString(2, infoboxContent);
			st.setInt(3, pageId);
			
			try {
				st.execute();
			}  catch (MysqlDataTruncation e1) {
				System.err.println("Infobox type too long: " + infoboxType);
			}
			st.close();
		}  catch (SQLException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Inserts found redirect to the database.
	 * @param oldName redirecting from old name
	 * @param newName redirecting to new name
	 */
	private void insertRedirectTable(String oldName, String newName) {
		PreparedStatement st;
		try {
			String query = ("INSERT INTO `wikipedia`.`redirect`  (`oldName`, `newName`) VALUES (?, ?)");
			st = databaseHandler.getConnection().prepareStatement(query);
			
			st.setString(1, oldName);
			st.setString(2, newName);

			try {
				st.execute();
			}  catch (SQLIntegrityConstraintViolationException e1) {
//				System.out.println("Redirect " + oldName + " -> " + newName + " is already in the database!");
			}

			st.close();
		} catch (SQLException e) {
			e.printStackTrace();
		}
	}
}
