package datahandlers;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.jgrapht.Graph;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.joda.time.DateTime;
import org.joda.time.DateTimeConstants;

/**
 * Class for handling directed graph.
 * @author Karel Šilhavý
 * @author Gabriela Hessová
 *
 */
public class JgraphtDataHandler {
	
	public static final String NUMBERS_SEPARATOR = "(\\D)+";
	public static final String AGED = "\\([aA]ged [\\d]*\\)";
	
	public static final String BC = "BC";
	public static final String BCE = "BCE";
	public static final String OR = "OR";
	
	public static final String[] monthNames = {"January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December"};
	
	private DatabaseHandler databaseHandler;
	
	private Graph<GraphVertex, RelationshipEdge> directedGraph;
	
	private int currentStateIndex;
	
	public List<LinkedHashSet<GraphVertex>> verticesHistory;
	
	private List<LinkedHashSet<StoredRelationshipEdge>> edgesHistory;
		
	public JgraphtDataHandler(DatabaseHandler databaseHandler) {
		this.databaseHandler = databaseHandler;
		this.directedGraph = new DefaultDirectedGraph<>(RelationshipEdge.class);
		this.verticesHistory = new ArrayList<LinkedHashSet<GraphVertex>>();
		this.edgesHistory = new ArrayList<LinkedHashSet<StoredRelationshipEdge>>();
		this.currentStateIndex = 0;
	}
	
	public void insertInfoboxIntoGraph(String infoboxName) throws SQLException {
		InfoboxEntry infobox = databaseHandler.getInfobox(infoboxName);
		if (infobox == null) { //page doesn't have infobox
			directedGraph.addVertex(new GraphVertex(infoboxName));			
		}
		else { //page has infobox
			DateTime birthDate = getBirthDate(infobox);
			DateTime deathDate = getDeathDate(infobox);
			directedGraph.addVertex(new GraphVertex(infoboxName, infobox.getFullHtmlPageString(), birthDate, deathDate));
		}
	}
	
	public DateTime getBirthDate(InfoboxEntry infobox) {
		for (int i = 0; i < infobox.getKeys().size(); i++) {
			String key = infobox.getKeys().get(i).getKey();
			if (key.equals("birth_date")) {
				return getDate(infobox.getLinks().get(i).get(0).getTitleValue());
			}
		}
		return null;
	}
	
	public DateTime getDeathDate(InfoboxEntry infobox) {
		for (int i = 0; i < infobox.getKeys().size(); i++) {
			String key = infobox.getKeys().get(i).getKey();
			if (key.equals("death_date")) {
				return getDate(infobox.getLinks().get(i).get(0).getTitleValue());
			}
		}
		return null;
	}
	
	/**
	 * @author Gabriela Hessová
	 * @param value date as string
	 * @return DateTime instance
	 */
	public static DateTime getDate(String value) {
		boolean beforeChrist = false;
		boolean found = false;
		int year = -1;
		int month = -1;
		int day = -1;
		
		if (value == null || value == "") {
			return null;
		}
		if (value.toUpperCase().contains(BC) || value.toUpperCase().contains(BCE)) {
			beforeChrist = true;
		}
		value = value.replaceAll(AGED, "");		
		String[] splitResult = value.split(NUMBERS_SEPARATOR);
		if (splitResult.length < 1) {
			return null;
		}
		ArrayList<Integer> numbers = new ArrayList<Integer>();
		int number;
		for (int i = 0; i < splitResult.length; i++) {
			try {
				number = Integer.parseInt(splitResult[i]);
				numbers.add(number);
			} catch (NumberFormatException e) {
				continue;
			}
		}
		if (numbers.isEmpty()) {
			return null;
		}
		if (numbers.size() == 1 && value.contains("1st")) {
			return null;
		}
		if (numbers.size() > 2) {
			year = numbers.get(0);
			month = numbers.get(1);
			day = numbers.get(2);
		}
		else if (numbers.size() == 1) {
			year = numbers.get(0);
			day = 1;
			month = 1;
		}
		
		if (value.contains(year + "|" + month + "|" + day) 
				|| value.contains(year + "|0" + month + "|" + day)
				|| value.contains(year + "|" + month + "|0" + day)
				|| value.contains(year + "|0" + month + "|0" + day)) {
			found = true;
		}
		
		if (!found) {
			
			Character c;
			String monthName = "";
			int monthNumber = -1;
			for (int i = 0; i < value.length(); i++) {
				c = value.charAt(i);

				if (Character.isUpperCase(c)) {
					monthName += c;
					for (int j = i + 1; j < value.length(); j++) {
						c = value.charAt(j);
						if (Character.isAlphabetic(c)) {
							monthName += c;
						} else { //není písmeno
							if ((monthNumber = getMonthNumber(monthName)) != -1) {
								i = value.length(); //skok na konec vnějšího cyklu                        		
							} else {
								monthName = "";
								i = j;
							}
							break; //konec vnitřního cyklu
						}
					}
				}
			}
			//nalezen měsíc
			year = numbers.get(0);
			if (monthNumber != -1) {
				month = monthNumber;
				if (numbers.size() > 1) {
					day = numbers.get(1);
					found = true;
				}
			} else {

			}
		}
		if (!found && value.toUpperCase().contains(OR)) {
			year = numbers.get(0);
			day = 1;
		}
		
		if (day > 31) {			
			int tmp = day;
			day = year;
			year = tmp;
			
		}
		if (day == -1 ) {
			day = 1;
		}
		if (month == -1 ) {
			month = 1;
		}
		if (month < 1 || month > 12) {
			month = 1;
			day = 1;
		}
		
		DateTime datetime = null;
		try {
			datetime = new DateTime(year, month, day, 0, 1);
		} catch (Exception e) {
			datetime = new DateTime(1,1,1,0,1);
		}
		if (beforeChrist) {
			datetime = datetime.withEra(DateTimeConstants.BC);
		}
		return datetime;
	}
	
	/**
	 * @author Gabriela Hessová
	 * @param month
	 * @return
	 */
	public static int getMonthNumber(String month) {
		for (int i = 0; i < monthNames.length; i++) {
			if (month.toUpperCase().equals(monthNames[i].toUpperCase())) {
				return i + 1;
			}
		}
		return -1;
	}
		
	public GraphVertex getVertexByName(String vertexName) {
		for (GraphVertex graphVertex: directedGraph.vertexSet()) {
			if (graphVertex.getName().equals(vertexName)) {
				return graphVertex;
			}
		}
		return null;
	}
	
	public void insertEdge(GraphVertex sourceVertex, GraphVertex targetVertex, String edgeName) {
		RelationshipEdge edge = new RelationshipEdge(edgeName);
		directedGraph.addEdge(sourceVertex, targetVertex, edge);
	}
	
	public void saveCurrentData() {
		LinkedHashSet<GraphVertex> currentVertices = new LinkedHashSet<GraphVertex>();
		for (GraphVertex currentGraphVertex: getAllVertices()) {
			currentVertices.add(new GraphVertex(currentGraphVertex.getName(), currentGraphVertex.getBirthDate(), currentGraphVertex.getDeathDate()));
		}
		verticesHistory.add(currentStateIndex, currentVertices);
		
		LinkedHashSet<StoredRelationshipEdge> currentEdges = new LinkedHashSet<StoredRelationshipEdge>();
		for (RelationshipEdge currentEdge: getAllEdges()) {
			currentEdges.add(new StoredRelationshipEdge(currentEdge.getLabel(), getEdgeSource(currentEdge), getEdgeTarget(currentEdge)));
		}
		edgesHistory.add(currentStateIndex, currentEdges);
		
		currentStateIndex++;
	}
	
	public void loadDataBefore() {
		if (currentStateIndex > 1) {
			currentStateIndex -= 2;
			loadStateOfcurrentIndex();
			currentStateIndex++;
		}
	}
	
	public void loadDataAfter() {
		if (verticesHistory.size() > currentStateIndex) {
			loadStateOfcurrentIndex();
			currentStateIndex++;
		}
	}
	
	private void loadStateOfcurrentIndex() {
		clearGraph();
		
		//fill vertices
		for (GraphVertex vertex: verticesHistory.get(currentStateIndex)) {
			directedGraph.addVertex(vertex);
		}
		
		//fill edges
		for (StoredRelationshipEdge edge: edgesHistory.get(currentStateIndex)) {
			directedGraph.addEdge(edge.getSourceVertex(), edge.getTargetVertex(), new RelationshipEdge(edge.getLabelOfRelationshipEdge()));
		}
	}
	
	public Set<GraphVertex> getAllVertices() {
		return directedGraph.vertexSet();
	}
	
	public Set<RelationshipEdge> getAllEdges() {
		return directedGraph.edgeSet();
	}

	public Set<RelationshipEdge> getEdgesOfVertex(String vertexName) {
		return directedGraph.edgesOf(getVertexByName(vertexName));
	}
	
	public GraphVertex getEdgeSource(RelationshipEdge edge) {
		return directedGraph.getEdgeSource(edge);
	}
	
	public GraphVertex getEdgeTarget(RelationshipEdge edge) {
		return directedGraph.getEdgeTarget(edge);
	}
	
	public void clearGraph() {
		directedGraph = new DefaultDirectedGraph<>(RelationshipEdge.class);
	}

	public void deleteVertex(String vertexToDelete) {
		directedGraph.removeVertex(getVertexByName(vertexToDelete));
	}

	public void deleteEdge(String nameOfEdgeToDelete) {
		for (RelationshipEdge edge: getAllEdges()) {
			if (edge.getLabel().equals(nameOfEdgeToDelete)) {
				directedGraph.removeEdge(edge);
				return;
			}
		}
	}
	
}
